/*
 * Copyright (C) 2002 2003 2004, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <config.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include "gtkfiles.h"
#include "um.h"
#include "chunk.h"
#include "inifile.h"
#include "main.h"
#include "ringbuf.h"
#include "rawdialog.h"
#include "tempfile.h"


static GtkObjectClass *parent_class;

static GList *chunks = NULL;

guint chunk_alive_count(void)
{
     return g_list_length ( chunks );
}

static void chunk_destroy(GtkObject *object)
{
     Chunk *c;
     GList *l;
     DataPart *dp;
     c=CHUNK(object);
     g_assert(c->opencount == 0);
     if (c->parts) {
	  for (l=c->parts; l!=NULL; l=l->next) {
	       dp = (DataPart *)l->data;
	       gtk_object_unref(GTK_OBJECT(dp->ds));
	       g_free(dp);
	  }
	  g_list_free(c->parts);
     }
     c->parts = NULL;
     parent_class->destroy(object);
     chunks = g_list_remove ( chunks, object );
}

static void chunk_class_init(GtkObjectClass *klass)
{
     parent_class = gtk_type_class(gtk_object_get_type());
     klass->destroy = chunk_destroy;
}

static void chunk_init(Chunk *object)
{
     object->format.samplerate=44100;
     object->format.samplesize=2;
     object->format.channels=2;
     object->format.sign=1;
     object->format.samplebytes=4;
     object->length=0;
     object->size=0;
     object->parts = 0;
     chunks = g_list_append( chunks, object );
}

guint chunk_get_type(void)
{
     static guint id=0;
     if (!id) {
	  GtkTypeInfo info = {
	       "Chunk",
	       sizeof(Chunk),
	       sizeof(ChunkClass),
	       (GtkClassInitFunc) chunk_class_init,
	       (GtkObjectInitFunc) chunk_init 
	  };
	  id=gtk_type_unique(gtk_object_get_type(),&info);
     }
     return id;
}

static Chunk *chunk_new(void)
{
     return gtk_type_new(chunk_get_type());
}

Chunk *chunk_new_silent(Dataformat *format, gfloat seconds)
{
     return chunk_new_from_datasource(
	  datasource_new_silent(format, seconds * format->samplerate));
}

static void chunk_calc_length(Chunk *c)
{
     GList *l;
     DataPart *dp;
     off_t len=0;
     for(l=c->parts; l!=NULL; l=l->next) {
	  dp = (DataPart *)l->data;
	  len += dp->length;
     }
     c->length = len;
     c->size = len*c->format.samplebytes;
}

static gint target_samplerate;
static gint samplecount;

gboolean chunk_convert_samplerate_proc(void *in, guint samplesize, 
				       chunk_writeout_func out_func, WriteoutID id,
				       Chunk *chunk)
{
     while (samplecount<target_samplerate) {
	  if (out_func(id,in,samplesize)) return TRUE;
	  samplecount += chunk->format.samplerate;
     }
     samplecount -= target_samplerate;
     return FALSE;
}

static sample_t last_samples[16];
static gfloat step_count;
static gfloat step_inc;

gboolean chunk_convert_samplerate_linear_proc(void *in, guint samplesize, 
					      chunk_writeout_func out_func, WriteoutID id,
					      Chunk *chunk)
{
     sample_t *f = (sample_t *)in, inc[16];
     gint i,steps;
     if (!samplecount) { 
	  samplecount++; 
	  memcpy(last_samples,f,samplesize);
	  return out_func(id,in,samplesize);
     }
     step_count += step_inc;
     steps = (gint)step_count;
     step_count -= (gfloat)steps;
     for (i=0; i<chunk->format.channels; i++) inc[i] = (f[i] - last_samples[i]) / steps;
     for (; steps>0; steps--) {
	  for (i=0; i<chunk->format.channels; i++)
	       last_samples[i] += inc[i];
	  if (out_func(id,last_samples,samplesize)) return TRUE;
     }
     return FALSE;
}

Chunk *chunk_convert_samplerate(Chunk *chunk, gint samplerate, 
				ChunkInterpolation interpolation, StatusBar *bar)
{
     Chunk *c;
     GList *l;
     DataPart *dp;
     target_samplerate = samplerate;
     samplecount = 0;
     step_count = 0.0;
     step_inc = (gfloat)samplerate / (gfloat)chunk->format.samplerate;
     if (interpolation == INTERPOLATION_LINEAR && step_inc > 1.0)
	  c = chunk_filter( chunk, chunk_convert_samplerate_linear_proc, 
			    NULL, CHUNK_FILTER_FULL, TRUE, bar, 
			    "Converting");
     else 
	  c = chunk_filter(chunk,chunk_convert_samplerate_proc,NULL,
			   CHUNK_FILTER_FULL,FALSE,bar,"Converting");
     if (c) {
	  c->format.samplerate = samplerate;
	  for (l=c->parts; l!=NULL; l=l->next) {
	       dp = (DataPart *)l->data;
	       dp->ds->format.samplerate = samplerate;
	  }
     }
     return c;
}

Chunk *chunk_convert_speed(Chunk *chunk, gfloat speed_factor, StatusBar *bar)
{
     target_samplerate = ((float)chunk->format.samplerate) / speed_factor;
     samplecount = 0;
     step_count = 0.0;
     step_inc = (gfloat)(chunk->format.samplerate) / (gfloat)chunk->format.samplerate;
     if (step_inc > 1.0) 
	  return chunk_filter(chunk,chunk_convert_samplerate_linear_proc,NULL,
			      CHUNK_FILTER_FULL,TRUE,bar,"Adjusting speed");
     else
	  return chunk_filter(chunk,chunk_convert_samplerate_proc,NULL,
			      CHUNK_FILTER_FULL,FALSE,bar,
			      "Adjusting speed");
}

static Chunk *chunk_convert_samplesize_part(DataPart *part, Dataformat *tfmt, 
					    StatusBar *bar)
{
     TempFile *tf;
     sample_t sbuf[1024];
     gpointer obuf;
     off_t p;
     guint i;
     guint samps;

     if (datasource_open(part->ds)) return NULL;
     tf = tempfile_init(tfmt,FALSE,FALSE);
     if (tf == NULL) { datasource_close(part->ds); return NULL; }
     obuf = g_malloc(ARRAY_LENGTH(sbuf)*tfmt->samplesize);
     p = part->position;
     while (p < (part->position+part->length)) {
	  samps = ARRAY_LENGTH(sbuf)/tfmt->channels;
	  if ((off_t)samps > part->position+part->length-p)
	       samps = (int)(part->position+part->length-p);
	  i = datasource_read_array_fp(part->ds,p,samps,sbuf);
	  convert_float_pcm_array(sbuf,obuf,tfmt,i*tfmt->channels);
	  if (i==0 || tempfile_write(tf,obuf,i*tfmt->samplebytes) ||
	      status_bar_progress(bar, i)) {
	       g_free(obuf);
	       datasource_close(part->ds);
	       tempfile_abort(tf);
	       return NULL;
	  }
	  p += i;
     }
     g_free(obuf);
     datasource_close(part->ds);
     return tempfile_finished(tf);
}

Chunk *chunk_convert_samplesize(Chunk *chunk, guint samplesize, gboolean sign,
				StatusBar *bar)
{
     GList *l;
     Datasource *ds;
     Chunk *r,*c,*c2;
     DataPart *dp;

     r = chunk_new();
     memcpy(&(r->format),&(chunk->format),sizeof(Dataformat));
     r->format.samplesize = samplesize;
     r->format.samplebytes = r->format.channels * samplesize;
     r->format.sign = sign;

     status_bar_begin_progress(bar,chunk->length,"Converting");
     for (l=chunk->parts; l!=NULL; l=l->next) {
	  dp = (DataPart *)(l->data);
	  ds = datasource_try_convert_samplesize(dp->ds,r->format.samplesize,
						 r->format.sign);
	  if (ds == NULL) {
	       c = chunk_convert_samplesize_part(dp,&(r->format),bar);
	       g_assert(c->length == dp->length);
	       if (c == NULL) {
		    gtk_object_sink(GTK_OBJECT(r));
		    return NULL;
	       }
	  } else {
	       if (status_bar_progress(bar,dp->length)) {
		    gtk_object_sink(GTK_OBJECT(ds));
		    gtk_object_sink(GTK_OBJECT(r));
		    return NULL;
	       } 
	       c = chunk_new_from_datasource(ds);	       
	  }
	  c2 = r;
	  r = chunk_append(c2,c);
	  gtk_object_sink(GTK_OBJECT(c2));
	  gtk_object_sink(GTK_OBJECT(c));
     }
     
     return r;
}

static gint channel_to_remove, current_channel;

static gboolean chunk_remove_channel_proc(void *in, guint samplesize, 
					  chunk_writeout_func out_func, WriteoutID id,
					  Chunk *chunk)
{
     guint c = current_channel;
     current_channel++;
     if (current_channel == chunk->format.channels) current_channel = 0;
     if (c != channel_to_remove) return out_func(id,in,samplesize);
     else return FALSE;
}

Chunk *chunk_remove_channel(Chunk *chunk, gint channel, StatusBar *bar)
{
     Chunk *c;
     GList *l;
     DataPart *dp;
     channel_to_remove = channel;
     current_channel = 0;
     c = chunk_filter(chunk,chunk_remove_channel_proc,NULL,
		      CHUNK_FILTER_ONE,FALSE,bar,"Removing channel");
     if (c) {
	  c->format.channels--;
	  c->format.samplebytes = c->format.samplesize * c->format.channels;
	  c->length = c->size / c->format.samplebytes;
	  for (l=c->parts; l!=NULL; l=l->next) {
	       dp = (DataPart *)l->data;
	       memcpy(&(dp->ds->format),&(c->format),sizeof(Dataformat));
	       dp->ds->length = dp->length = 
		    dp->ds->size / dp->ds->format.samplebytes;
	  }
     }
     return c;
}

gboolean chunk_dump(Chunk *chunk, EFILE *file, gboolean bigendian, 
		    StatusBar *bar)
{
     GList *l;
     DataPart *dp;
     l = chunk->parts;
     while (l) {
	  dp = (DataPart *)l->data;
	  if (datasource_dump(dp->ds,dp->position,dp->length,file,bigendian,
			      bar))
	       return TRUE;
	  l = l->next;
     }
     return FALSE;
}

static Chunk *chunk_filter_chunk;

static gboolean chunk_filter_sample_t_write(WriteoutID id, gpointer data, guint length)
{
     static sample_t buf[512];
     static guint bufpos=0;
     sample_t *d = (sample_t *)data;
     gchar buf2[2048];
     guint i, count;
     if (!data) {
	  if (bufpos == 0) return FALSE;
	  i = bufpos;
	  bufpos = 0;
	  chunk_sample_put_array(buf,buf2,i,chunk_filter_chunk);
	  return tempfile_write(id,buf2,
				i*chunk_filter_chunk->format.samplesize);
     }
     count = length / sizeof(sample_t);
     while (count > 0) {
	  i = ARRAY_LENGTH(buf)-bufpos;
	  if (i>count) i=count;
	  memcpy(buf+bufpos,d,i*sizeof(sample_t));
	  bufpos += i;
	  if (bufpos == ARRAY_LENGTH(buf)) {
	       chunk_sample_put_array(buf,buf2,bufpos,chunk_filter_chunk);
	       if (tempfile_write(id,buf2,bufpos*
				  chunk_filter_chunk->format.samplesize))
		    return TRUE;
	       bufpos = 0;
	  }
	  d += i;
	  count -= i;
     }
     return FALSE;
}

static gboolean has_fake_pcm(Chunk *c)
{
     GList *l;
     DataPart *part;
     for (l=c->parts; l!=NULL; l=l->next) {
	  part = (DataPart *)(l->data);
	  if (part->ds->fake_pcm) return TRUE;
     }
     return FALSE;
}

/* Allmn funktion fr stegvis bearbetning av en Chunk.
 * Funktionen lser steg fr steg av chunken och anropar proc.  
 * Efter allt data skickats in kommer eof_proc (om != NULL) att anropas med 
 * samplesize=0.
 */

Chunk *chunk_filter(Chunk *chunk, chunk_filter_proc proc, 
		    chunk_filter_proc eof_proc, 
		    gint amount, gboolean convert,
		    StatusBar *bar, gchar *title)
{
     guint proc_size,single_size,full_size;
     gchar buf[8192];
     off_t samplepos=0, samplesleft=chunk->length;
     guint u, x;
     gboolean fake_pcm;
     ChunkHandle *ch;
     TempFile tmp;
     Chunk *ds;     

     chunk_filter_chunk = chunk;

     if (convert) single_size = sizeof(sample_t);
     else single_size = chunk->format.samplesize;
     
     full_size = single_size * chunk->format.channels;

     if (amount == CHUNK_FILTER_FULL) proc_size = full_size;     
     else if (amount == CHUNK_FILTER_ONE) proc_size = single_size;
     else proc_size = 0; /* Set this later for CHUNK_FILTER_MANY */

     /* If the fake_pcm flag is set for any of the Datasources we're
      * processing, use fake pcm output */ 
     fake_pcm = has_fake_pcm(chunk);     

     ch = chunk_open(chunk);
     if (ch == NULL) return NULL;
     tmp = tempfile_init(&(chunk->format),fake_pcm,FALSE);
     status_bar_begin_progress(bar,samplesleft,title);

     while (samplesleft > 0) {
	  if (convert)
	       u = chunk_read_array_fp(ch,samplepos,
				       MIN(sizeof(buf)/full_size,samplesleft),
				       (sample_t *)buf) * full_size;
	  else
	       u = chunk_read_array(ch,samplepos,
				    MIN(sizeof(buf),samplesleft*full_size),buf);
	  if (!u) {
	       chunk_close(ch);
	       tempfile_abort(tmp);
	       return NULL;
	  }
	  if (amount == CHUNK_FILTER_MANY) proc_size = u;
	  for (x=0; x<u; x+=proc_size) {
	       if (proc(buf+x,proc_size,(convert && !fake_pcm) ?
			chunk_filter_sample_t_write:
			tempfile_write,tmp,
			chunk)) {
			 chunk_close(ch);
			 tempfile_abort(tmp);
			 return NULL;
	       }
	  }
	  samplesleft -= u / full_size;
	  samplepos += u / full_size;
	  if (status_bar_progress(bar, u / full_size)) {
	       chunk_close(ch);
	       tempfile_abort(tmp);
	       return NULL;
	  }
     }
     if (eof_proc && eof_proc(NULL,0,
			      (convert && !fake_pcm) ? 
			      chunk_filter_sample_t_write :
			      tempfile_write,tmp,chunk)) {	  
	  chunk_close(ch);
	  tempfile_abort(tmp);
	  return NULL;
     }
     if (convert && !fake_pcm && chunk_filter_sample_t_write(tmp,NULL,0)) {
	  chunk_close(ch);
	  tempfile_abort(tmp);
	  return NULL;
     }
     chunk_close(ch);
     ds = tempfile_finished(tmp);
     return ds;
}

/* Allmn funktion fr stegvis avlsning av en Chunk. proc kommer att anropas 
 * en gng fr varje sampling i chunken */

gboolean chunk_parse(Chunk *chunk, chunk_parse_proc proc, 
		     gboolean allchannels, gboolean convert, 
		     StatusBar *bar, gchar *title)
{

     guint single_size,full_size,proc_size;
     gchar buf[4096],*c=NULL,*d;
     off_t samplepos=0, samplesleft=chunk->length;
     guint u, x;
     ChunkHandle *ch;

     if (convert) single_size = sizeof(sample_t);
     else single_size = chunk->format.samplesize;
     
     full_size = single_size * chunk->format.channels;

     if (allchannels) proc_size = full_size;
     else proc_size = single_size;


     ch = chunk_open(chunk);
     if (ch == NULL) { g_free(c); return TRUE; }
     status_bar_begin_progress( bar, chunk->length, title);

     while (samplesleft > 0) {
	  if (convert)	       
	       u = chunk_read_array_fp(ch,samplepos,sizeof(buf)/full_size,
				       (sample_t *)buf) * full_size;
	  else
	       u = chunk_read_array(ch,samplepos,sizeof(buf),buf);
	  if (!u) {
	       chunk_close(ch);
	       g_free(c);
	       return TRUE;
	  }
	  d = buf;
	  for (x=0; x<u; x+=proc_size) {
	       if (proc(d+x,proc_size,chunk)) {
		    chunk_close(ch);
		    g_free(c);
		    return TRUE;
	       }
	  }
	  samplesleft -= u / full_size;
	  samplepos += u / full_size;
	  if (status_bar_progress(bar, u / full_size)) {
	       chunk_close(ch);
	       g_free(c);
	       return TRUE;
	  }
     }
     chunk_close(ch);
     g_free(c);
     return FALSE;
}

static gboolean chunk_onechannel_proc(void *in, guint sample_size, 
				      chunk_writeout_func out_func, 
				      WriteoutID id, Chunk *chunk)
{
     static int channel_count = 0;
     static sample_t mixup = 0.0;
     sample_t s = *((sample_t *)in);
     mixup += s;
     channel_count++;
     if (channel_count == chunk->format.channels) {
	  s = mixup / (sample_t)channel_count;
	  channel_count = 0;
	  mixup = 0.0;
	  return out_func(id,&s,sizeof(s));
     } else 
	  return FALSE;
}

Chunk *chunk_onechannel(Chunk *chunk, StatusBar *bar)
{
     Chunk *c;
     GList *l;
     DataPart *dp;

     c = chunk_filter(chunk,chunk_onechannel_proc,NULL,
		      CHUNK_FILTER_ONE,TRUE,bar,NULL);
     if (c) {
	  c->format.channels = 1;
	  c->format.samplebytes = c->format.samplesize;
	  c->length = c->size / c->format.samplebytes;
	  for (l=c->parts; l!=NULL; l=l->next) {
	       dp = (DataPart *)(l->data);
	       memcpy(&(dp->ds->format),&(c->format),sizeof(Dataformat));
	       dp->ds->length = dp->length = 
		    dp->ds->size/dp->ds->format.samplebytes;
	  }
     }
     return c;
}

static gint channel_to_copy;

static gboolean chunk_copy_channel_proc(void *in, guint sample_size, 
					chunk_writeout_func out_func, 
					WriteoutID id,
					Chunk *chunk)
{
     static gint channel_count=0;
     gint i;
     i = channel_count;
     channel_count ++;
     if (channel_count == chunk->format.channels) channel_count = 0;
     if (out_func(id,in,sample_size)) return TRUE;
     if (channel_to_copy==i) return out_func(id,in,sample_size);
     else return FALSE;
}

static gboolean chunk_copy_channel_mono_proc(void *in, guint sample_size,
					     chunk_writeout_func out_func, WriteoutID id,
					     Chunk *chunk)
{
     return (out_func(id,in,sample_size) || out_func(id,in,sample_size));
}

Chunk *chunk_copy_channel(Chunk *chunk, gint channel, StatusBar *bar)
{
     Chunk *c;
     GList *l;
     DataPart *dp;
     channel_to_copy = channel;
     if (chunk->format.channels == 1) 
	  c = chunk_filter(chunk,chunk_copy_channel_mono_proc,NULL,
			   CHUNK_FILTER_ONE,FALSE,bar,NULL);
     else 
	  c = chunk_filter(chunk,chunk_copy_channel_proc,NULL,
			   CHUNK_FILTER_ONE,FALSE,bar,NULL);
     if (c) {
	  c->format.channels++;
	  c->format.samplebytes = c->format.samplesize * c->format.channels;
	  c->length = c->size / c->format.samplebytes;
	  for (l=c->parts; l!=NULL; l=l->next) {
	       dp = (DataPart *)(l->data);
	       memcpy(&(dp->ds->format),&(c->format),sizeof(Dataformat));
	       dp->length = dp->ds->length = 
		    dp->ds->size / dp->ds->format.samplebytes;
	  }
     }
     return c;
}


Chunk *chunk_mix(Chunk *c1, Chunk *c2, StatusBar *bar)
{
     #define BUFLEN 256
     off_t u,mixlen;
     guint i,chn,x1,x2,x;
     guint clipcount = 0;
     sample_t buf1[BUFLEN],buf2[BUFLEN],buf3[BUFLEN];
     sample_t s,pmax;
     gpointer outbuf;
     gchar *str;
     gboolean fake_pcm;
     Chunk *m,*c,*d;
     ChunkHandle *ch1, *ch2;
     TempFile tmp;

     g_assert(chunk_format_equal(c1,c2));

     pmax = maximum_float_value(&(c1->format));

     /* Number of channels */
     chn = c1->format.channels;
     /* Number of multi-channel samples to mix */
     mixlen = MIN(c1->length, c2->length);

     fake_pcm = has_fake_pcm(c1) || has_fake_pcm(c2);

     /* Prepare for processing */
     ch1 = chunk_open(c1);
     if ( ch1 == NULL ) return NULL;
     ch2 = chunk_open(c2);
     if ( ch2 == NULL ) { chunk_close(ch1); return NULL; }
     if (!fake_pcm)
	  outbuf = g_malloc(BUFLEN * c1->format.samplesize);
     else
	  outbuf = NULL;
     tmp = tempfile_init(&(c1->format),fake_pcm,FALSE);
     status_bar_begin_progress(bar, mixlen, "Mixing");

     for (u=0; u<mixlen; u+=x) {

	  x1 = chunk_read_array_fp(ch1,u,BUFLEN/chn,buf1);
	  x2 = chunk_read_array_fp(ch2,u,BUFLEN/chn,buf2);
	  if (x1 == 0 || x2 == 0) {
	       tempfile_abort(tmp);
	       chunk_close(ch1);
	       chunk_close(ch2);
	       g_free(outbuf);
	       return NULL;
	  }

	  /* Number of mixable multi-channel samples for this run */
	  x = MIN(x1,x2);
	  /* Mix the data from buf1,buf2 into buf3 */
	  for (i=0; i<x*chn; i++) {
	       s = buf1[i] + buf2[i];
	       if (s > pmax) { clipcount ++; s = pmax; }
	       else if (s < -1.0) { clipcount ++; s = -1.0; }
	       buf3[i] = s;
	  }	  
	  if (fake_pcm) {
	       /* Write buf3 directly */
	       if (tempfile_write(tmp,buf3,x*chn*sizeof(sample_t))
		   || status_bar_progress(bar,x)) {
		    tempfile_abort(tmp);
		    chunk_close(ch1);
		    chunk_close(ch2);
		    g_free(outbuf);
		    return NULL;
	       }
	  } else {
	       /* Convert buf3 -> outbuf and write out*/
	       convert_float_pcm_array(buf3, outbuf, &(c1->format), x*chn);
	       if (tempfile_write(tmp,outbuf,x*c1->format.samplebytes)
		   || status_bar_progress(bar,x)) {
		    tempfile_abort(tmp);
		    chunk_close(ch1);
		    chunk_close(ch2);
		    g_free(outbuf);
		    return NULL;
	       }
	  }
     }

     /* Finish processing */
     chunk_close(ch1);
     chunk_close(ch2);
     m = tempfile_finished(tmp);
     g_free(outbuf);
     if (!m) return NULL;

     /* Warn if clipping occured */
     if (clipcount > 0) {
	  str = g_strdup_printf("The mixed result was clipped %d times.",
				clipcount);
	  user_warning(str);
	  g_free(str);
     }

     /* We may need to append some unmixed data at the end */
     if (c1->length == c2->length) return m;
     if (c1->length > mixlen)
	  c = chunk_get_part(c1,mixlen,c1->length-mixlen);
     else 
	  c = chunk_get_part(c2,mixlen,c2->length-mixlen);
     d = chunk_append(m,c);
     gtk_object_sink(GTK_OBJECT(m));
     gtk_object_sink(GTK_OBJECT(c));

     return d;
}


void chunk_foreach(GFunc func, gpointer user_data)
{
     g_list_foreach(chunks,func,user_data);
}

Chunk *chunk_new_from_datasource(Datasource *ds)
{
     Chunk *c;
     DataPart *dp;
     if (ds==NULL) return NULL;
     dp = g_malloc(sizeof(*dp));
     dp->ds = ds;
     dp->position = 0;
     dp->length = ds->length;
     gtk_object_ref(GTK_OBJECT(ds));
     gtk_object_sink(GTK_OBJECT(ds));
     c = chunk_new();
     memcpy(&(c->format),&(ds->format),sizeof(Dataformat));
     c->length = ds->length;
     c->size = c->length * c->format.samplebytes;
     c->parts = g_list_append(NULL,dp);
     return c;
}

ChunkHandle *chunk_open(Chunk *chunk)
{
     GList *l;
     DataPart *dp;
     for (l=chunk->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (datasource_open(dp->ds)) {
	       for (l=l->prev;l!=NULL;l=l->prev) {
		    dp = (DataPart *)l->data;
		    datasource_close(dp->ds);
	       }
	       return NULL;
	  }
     }
     chunk->opencount ++;
     return chunk;     
}

void chunk_close(ChunkHandle *handle)
{
     GList *l;
     DataPart *dp;
     g_assert(handle->opencount > 0);
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  datasource_close(dp->ds);
     }
     handle->opencount --;     	  
}

gboolean chunk_read(ChunkHandle *handle, off_t sampleno, void *buffer)
{
     GList *l;     
     DataPart *dp;
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) 
	       return datasource_read(dp->ds,sampleno,buffer);
	  else
	       sampleno -= dp->length;
     }
     g_assert_not_reached();
     return TRUE;
}

gboolean chunk_read_fp(ChunkHandle *handle, off_t sampleno, sample_t *buffer)
{
     GList *l;     
     DataPart *dp;
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) 
	       return datasource_read_fp(dp->ds,dp->position+sampleno,buffer);
	  else
	       sampleno -= dp->length;
     }
     g_assert_not_reached();
     return TRUE;
}

guint chunk_read_array(ChunkHandle *handle, off_t sampleno, 
		       guint size, void *buffer)
{
     GList *l;     
     DataPart *dp;
     guint r=0,i,j;
     off_t o;
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) {
	       /* Calculate required size if it's larger than we need. */
	       o = dp->length - sampleno;	       
	       o *= (off_t)(handle->format.samplebytes);
	       if (o < (off_t)size) j = (guint)o;
	       else j = size;
	       /* Read data. */
	       i = datasource_read_array(dp->ds,
					 dp->position + sampleno,
					 j,buffer);
	       if (i == 0) return 0;
	       r+=i;
	       sampleno = 0;
	       size -= i;
	       buffer = ((gchar *)buffer) + i;
	       if (size==0 || l->next==NULL) return r;
	  } else
	       sampleno -= dp->length;
     }
     return r;
}

guint chunk_read_array_fp(ChunkHandle *handle, off_t sampleno, 
			  guint samples, sample_t *buffer)
{
     GList *l;     
     DataPart *dp;
     guint r=0,i,j;
     off_t o;
     g_assert(sampleno < handle->length);
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) {
	       o = dp->length-sampleno;
	       if (o < (off_t)samples)
		    j = (guint)o;
	       else
		    j = samples;
	       i = datasource_read_array_fp(dp->ds,
					    dp->position+sampleno,
					    j,buffer);
	       g_assert(i <= samples);
	       g_assert(i <= dp->length - sampleno);
	       if (i==0) return 0;
	       r += i;
	       buffer += i*handle->format.channels;
	       samples -= i;
	       sampleno = 0;
	       if (samples == 0 || l->next==NULL) return r;
	  } else
	       sampleno -= dp->length;
     }
     return r;
}

static DataPart *datapart_list_copy(GList *src, GList **dest)
{
     DataPart *old,*new;
     old = src->data;
     new = g_malloc(sizeof(DataPart));
     new->ds = old->ds;
     new->position = old->position;
     new->length = old->length;
     gtk_object_ref(GTK_OBJECT(old->ds));
     (*dest) = g_list_append((*dest),new);
     return new;
}

Chunk *chunk_append(Chunk *first, Chunk *second)
{
     Chunk *c;
     GList *nl=NULL,*l;
     off_t len=0;
     DataPart *dp;
     g_assert(chunk_format_equal(first,second));
     for (l=first->parts; l!=NULL; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  len += dp->length;
     }
     for (l=second->parts; l!=NULL; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  len += dp->length;
     }
     c = chunk_new();
     memcpy(&(c->format),&(first->format),sizeof(Dataformat));
     c->parts = nl;
     c->length = len;
     c->size = len * c->format.samplebytes;
     return c;
}

Chunk *chunk_insert(Chunk *chunk, Chunk *part, off_t sampleno)
{
     GList *nl=NULL,*l,*m;     
     DataPart *dp;
     Chunk *c;

     g_assert(sampleno <= chunk->length);
     g_assert(chunk_format_equal(chunk,part));
     if (sampleno == chunk->length) return chunk_append(chunk,part);

     /* Get beginning */
     for (l=chunk->parts; sampleno>0; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  if (sampleno < dp->length) {
	       dp->length = sampleno;
	       break;
	  }
	  sampleno -= dp->length;
     }

     /* Add the inserted part */
     for (m=part->parts; m!=NULL; m=m->next) 
	  datapart_list_copy(m,&nl);

     /* Add the rest */
     for (; l!=NULL; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  dp->position += sampleno;
	  dp->length -= sampleno;
	  sampleno = 0;
     }

     /* Create the chunk */
     c = chunk_new();
     memcpy(&(c->format),&(chunk->format),sizeof(Dataformat));
     c->parts = nl;
     c->length = chunk->length + part->length;
     c->size = chunk->size + part->size;
     return c;
}

Chunk *chunk_get_part(Chunk *chunk, off_t start, off_t length)
{
     GList *l,*nl=NULL;
     DataPart *dp;
     Chunk *c;
     l=chunk->parts;
     /* Skip beginning parts */
     while (1) {
	  dp = (DataPart *)l->data;
	  if (start < dp->length) break;
	  start -= dp->length;
	  l=l->next;
     }
     while (length>0 && l!=NULL) {
	  dp = datapart_list_copy(l,&nl);
	  dp->position += start;
	  dp->length -= start;
	  start = 0;
	  if (dp->length > length) dp->length=length;
	  length -= dp->length;
	  l=l->next;
     }

     c = chunk_new();
     memcpy(&(c->format),&(chunk->format),sizeof(Dataformat));
     c->parts = nl;
     chunk_calc_length(c);
     return c;
}

Chunk *chunk_remove_part(Chunk *chunk, off_t start, off_t length)
{
     GList *nl=NULL,*l;
     DataPart *dp;
     Chunk *c;
     
     /* Copy beginning parts */
     l=chunk->parts;
     while (start>0) {
	  dp = datapart_list_copy(l,&nl);
	  if (start < dp->length) {
	       dp->length = start;
	       break;
	  }
	  start -= dp->length;
	  l=l->next;
     }
     
     /* Skip parts */
     length += start;
     while (length>0 && l!=NULL) {
	  dp = (DataPart *)l->data;
	  if (dp->length > length) break;
	  length -= dp->length;
	  l=l->next;
     }

     /* Copy remaining parts */
     for (;l!=NULL;l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  dp->position += length;
	  dp->length -= length;	  
	  length = 0;
     }

     c = chunk_new();
     memcpy(&(c->format),&(chunk->format),sizeof(Dataformat));
     c->parts = nl;
     chunk_calc_length(c);
     return c;
}

Chunk *chunk_replace_part(Chunk *chunk, off_t start, off_t length, Chunk *new)
{
     Chunk *c;
     GList *nl=NULL,*l;
     DataPart *dp;
     guint i;
     for (l=new->parts; l!=NULL; l=l->next)
	  datapart_list_copy(l,&nl);
     c = chunk_remove_part(chunk,start,length);
     for (l=c->parts,i=0; start>0; l=l->next,i++) {
	  dp = (DataPart *)l->data;
	  g_assert(dp->length <= start);
	  start -= dp->length;
     }
     for (l=nl; l!=NULL; l=l->next)
	  c->parts = g_list_insert(c->parts, l->data, i);
     g_list_free(nl);
     c->length += new->length;
     c->size = c->length * c->format.samplebytes;
     return c;
}

Chunk *chunk_clone_df(Chunk *chunk, Dataformat *format, gboolean byteswap)
{
     Chunk *c;
     GList *l,*nl=NULL;
     DataPart *dp,*ndp;
     Dataformat fmt;

     memcpy(&fmt,format,sizeof(Dataformat));
     for (l=chunk->parts; l!=NULL; l=l->next) {
	  dp = (DataPart *)l->data;
	  ndp = g_malloc(sizeof(*ndp));
	  ndp->ds = dp->ds;
	  if (!dataformat_equal(format,&(chunk->format))) 
	       ndp->ds = datasource_clone_df(ndp->ds, &fmt);
	  if (byteswap) ndp->ds = datasource_byteswap(ndp->ds);
	  gtk_object_ref(GTK_OBJECT(ndp->ds));
	  gtk_object_sink(GTK_OBJECT(ndp->ds));
	  ndp->position = dp->position;
	  ndp->length = (dp->length * dp->ds->format.samplebytes) /
	       ndp->ds->format.samplebytes;
	  nl = g_list_append(nl,ndp);
     }

     c = chunk_new();
     memcpy(&(c->format),format,sizeof(Dataformat));
     c->parts = nl;
     chunk_calc_length(c);
     return c;
}


static sample_t chunk_peak_level_max;

static gboolean chunk_peak_level_proc(void *sample, gint sample_size, 
				      Chunk *chunk)
{
     sample_t s;
     s = (sample_t)fabs((double)(*((sample_t *)sample)));
     if (s > chunk_peak_level_max) {
	  chunk_peak_level_max = s;
	  if (s == 1.0) return TRUE;
     }
     return FALSE;
}

sample_t chunk_peak_level(Chunk *c, StatusBar *bar)
{
     chunk_peak_level_max = 0.0;
     if (chunk_parse(c,chunk_peak_level_proc,FALSE,TRUE,bar,
		     "Calculating peak level")) return -1.0;
     return chunk_peak_level_max;
}

static sample_t chunk_amplify_factor;

static gboolean chunk_amplify_proc(void *in, guint sample_size,
				   chunk_writeout_func out_func,
				   WriteoutID id, Chunk *source)
{
     sample_t *sp,*ep;
     sp = ((sample_t *)in);
     ep = (sample_t *)(((gchar *)in)+sample_size);
     for (; sp<ep; sp++) *sp *= chunk_amplify_factor;
     return out_func(id,in,sample_size);
}

Chunk *chunk_amplify(Chunk *c, sample_t factor, StatusBar *bar)
{
     chunk_amplify_factor = factor;
     return chunk_filter(c,chunk_amplify_proc,NULL,CHUNK_FILTER_MANY,TRUE,bar,
			 "Amplifying");
}

static sample_t ramp_start,ramp_diff;
static off_t samples_done;

static gboolean volume_ramp_proc(void *in, guint sample_size,
				 chunk_writeout_func out_func,
				 WriteoutID id, Chunk *source)
{
     sample_t *sp = (sample_t *)in;
     sample_t factor;
     int i,j;
     for (j=0; j<sample_size; j+=source->format.channels*sizeof(sample_t)) {
	  factor = ramp_start + (ramp_diff * ((sample_t)samples_done / 
					      (sample_t)(source->length)));
	  for (i=0; i<source->format.channels; i++) {
	       *sp *= factor;
	       sp++;
	  }
	  samples_done ++;
     }
     return out_func(id,in,sample_size);
}

Chunk *chunk_volume_ramp(Chunk *c, sample_t start_factor, sample_t end_factor, 
			 StatusBar *bar)
{
     if (start_factor == end_factor) return chunk_amplify(c,start_factor,bar);
     ramp_start = start_factor;
     ramp_diff = end_factor - start_factor;
     samples_done = 0;
     return chunk_filter(c,volume_ramp_proc,NULL,CHUNK_FILTER_MANY,TRUE,bar,
			 "Amplifying");
}

Chunk *chunk_interpolate_endpoints(Chunk *chunk, StatusBar *bar)
{
#define BUF_LENGTH 512
     sample_t startval[8],diffval[8];
     sample_t *sbuf;
     gpointer pcmbuf;
     off_t ctr=0,length=chunk->length;
     guint bufctr;
     guint i;
     sample_t slength = (sample_t)length;
     guint k,channels=chunk->format.channels;
     ChunkHandle *ch;
     TempFile tf;
     /* get the endpoint values */
     ch = chunk_open(chunk);
     if (ch == NULL) return NULL;
     if (chunk_read_fp(ch,0,startval) || 
	 chunk_read_fp(ch,length-1,diffval)) { 
	  chunk_close(ch);
	  return NULL;
     }
     chunk_close(ch);
     for (k=0;k<channels;k++) {
	  diffval[k]-=startval[k];
	  /* printf("%f %f\n",startval[k],diffval[k]); */
     }
     /* Prepare processing */
     tf = tempfile_init(&(chunk->format),FALSE,FALSE);
     if (tf == NULL) return NULL;
     status_bar_begin_progress(bar,chunk->length,NULL);
     sbuf = g_malloc(BUF_LENGTH*sizeof(sample_t));
     pcmbuf = g_malloc(BUF_LENGTH*chunk->format.samplesize);
     /* Output data */
     while (ctr < length) {
	  bufctr = BUF_LENGTH / channels;
	  if ((length-ctr) < (off_t)bufctr) bufctr = (guint)(length-ctr);
	  for (i=0; i<bufctr*channels; i+=channels,ctr++)
	       for (k=0; k<channels; k++)
		    sbuf[i+k] = startval[k] + 
			 diffval[k]*(((sample_t)ctr)/slength);
	  convert_float_pcm_array(sbuf,pcmbuf,&(chunk->format),
				  bufctr*channels);
	  if (tempfile_write(tf,pcmbuf,bufctr*chunk->format.samplebytes) ||
	      status_bar_progress(bar,bufctr)) {
	       g_free(pcmbuf);
	       g_free(sbuf);
	       tempfile_abort(tf);	       
	       return NULL;
	  }
     }
     /* Finish */
     g_free(pcmbuf);
     g_free(sbuf);
     return tempfile_finished(tf);
#undef BUF_LENGTH
}

Chunk *chunk_byteswap(Chunk *chunk)
{
     return chunk_clone_df(chunk,&(chunk->format),TRUE);
}

