/*
 * Copyright (C) 2002 2003 2004, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <config.h>

#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "gtkfiles.h"
#include "ringbuf.h"
#include "um.h"
#include "main.h"
#include "inifile.h"

#include "datasource.h"
#include "tempfile.h"

static GtkObjectClass *parent_class;
static GList *datasource_list = NULL;

guint datasource_count(void)
{
     return g_list_length(datasource_list);
}

static void datasource_init(Datasource *obj)
{
     datasource_list = g_list_append(datasource_list, obj);
     obj->type = DATASOURCE_SILENCE;
     obj->format.samplerate=44100;
     obj->format.samplesize=2;
     obj->format.channels=2;
     obj->format.sign=1;
     obj->format.samplebytes=4;
     obj->length = 0;
     obj->size = 0;
     obj->fake_pcm = FALSE;
     obj->opencount = 0;
     obj->tag = 0;
}

static void datasource_clear(Datasource *ds)
{
     g_assert(ds->opencount == 0);

     switch (ds->type) {

     case DATASOURCE_TEMPFILE:
     case DATASOURCE_TEMPFILE_FLOAT:
	  if (ds->data.virtual.filename)
	       xunlink(ds->data.virtual.filename);	  
	  /* Fall through */

     case DATASOURCE_VIRTUAL:
	  if (ds->data.virtual.filename) {
	       g_free(ds->data.virtual.filename);
	       ds->data.virtual.filename = NULL;
	  }
	  break;
	  
     case DATASOURCE_REAL:
     case DATASOURCE_REAL_FLOAT:
	  if (ds->data.real) {
	       g_free(ds->data.real);
	       ds->data.real = NULL;
	  }
	  break;

     case DATASOURCE_SNDFILE_TEMPORARY:
	  if (ds->data.sndfile.filename)
	       xunlink(ds->data.sndfile.filename);
	  /* Fall through */

     case DATASOURCE_SNDFILE:
	  if (ds->data.sndfile.filename) {
	       g_free(ds->data.sndfile.filename);
	       ds->data.sndfile.filename = NULL;
	  }
	  break;

     case DATASOURCE_REF:
     case DATASOURCE_CLONE:
     case DATASOURCE_BYTESWAP:
	  if (ds->data.clone) 
	       gtk_object_unref(GTK_OBJECT(ds->data.clone));
	  ds->data.clone = NULL;
     }
     
     ds->type = DATASOURCE_SILENCE;
}

static void datasource_destroy(GtkObject *obj)
{     
     Datasource *ds = DATASOURCE(obj);
     datasource_clear(ds);
     datasource_list = g_list_remove(datasource_list, obj);
     parent_class->destroy(obj);
}

static void datasource_class_init(GtkObjectClass *klass)
{
     parent_class = gtk_type_class(gtk_object_get_type());
     klass->destroy = datasource_destroy;
}

GtkType datasource_get_type(void)
{
     static GtkType id = 0;
     if (!id) {
		GtkTypeInfo info = {
		     "Datasource",
		     sizeof(Datasource),
		     sizeof(DatasourceClass),
		     (GtkClassInitFunc) datasource_class_init,
		     (GtkObjectInitFunc) datasource_init 
		};
		id=gtk_type_unique(gtk_object_get_type(),&info);	  
     }
     return id;
}

/* Creates a copy of the original datasource
 * If it refers to a file, create a DATASOURCE_REF instead.  
 */

static Datasource *datasource_copy(Datasource *orig)
{
     Datasource *ds;
     off_t u;

     ds = gtk_type_new(datasource_get_type());
     ds->type = orig->type;
     memcpy(&(ds->format),&(orig->format),sizeof(Dataformat));
     ds->length = orig->length;
     ds->size = orig->size;
     ds->fake_pcm = orig->fake_pcm;

     switch (orig->type) {
     case DATASOURCE_VIRTUAL:
     case DATASOURCE_TEMPFILE:
     case DATASOURCE_TEMPFILE_FLOAT:
     case DATASOURCE_SNDFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
	  ds->type = DATASOURCE_REF;	  
	  ds->data.clone = orig;
	  gtk_object_ref(GTK_OBJECT(orig));
	  gtk_object_sink(GTK_OBJECT(orig));
	  break;
     case DATASOURCE_REAL:
	  ds->data.real = g_malloc(orig->size);
	  memcpy(ds->data.real, orig->data.real, orig->size);
	  break;
     case DATASOURCE_REAL_FLOAT:
	  u = orig->length * orig->format.channels * sizeof(sample_t);
	  ds->data.real = g_malloc(u);
	  memcpy(ds->data.real, orig->data.real, u);
	  break;
     case DATASOURCE_CLONE:
     case DATASOURCE_BYTESWAP:
     case DATASOURCE_REF:
	  ds->data.clone = orig->data.clone;
	  gtk_object_ref(GTK_OBJECT(ds->data.clone));
	  gtk_object_sink(GTK_OBJECT(ds->data.clone));
     }
     return ds;
}

Datasource *datasource_clone_df(Datasource *source, Dataformat *format)
{
     Datasource *df;

     if (source->type == DATASOURCE_REAL ||
	 source->type == DATASOURCE_SILENCE ||
	 source->type == DATASOURCE_CLONE) {
	  df = datasource_copy(source);
	  g_assert(df->type == source->type);
	  memcpy(&(df->format),format,sizeof(Dataformat));
	  df->length = df->size / format->samplebytes;
	  return df;
     }
 
     df = gtk_type_new(datasource_get_type());
     df->type = DATASOURCE_CLONE;
     memcpy(&(df->format),format,sizeof(Dataformat));
     df->fake_pcm = FALSE;
     df->size = source->size;
     df->length = source->size / format->samplebytes;
     df->data.clone = source;
     gtk_object_ref(GTK_OBJECT(source));
     return df;
}

Datasource *datasource_byteswap(Datasource *source)
{
     Datasource *ds;
     if (source == NULL) return NULL;
     ds = gtk_type_new(datasource_get_type());
     ds->type = DATASOURCE_BYTESWAP;
     memcpy(&(ds->format),&(source->format),sizeof(Dataformat));
     ds->length = source->length;
     ds->size = source->size;
     ds->fake_pcm = FALSE;
     ds->data.clone = source;
     gtk_object_ref(GTK_OBJECT(source));
     gtk_object_sink(GTK_OBJECT(source));
     return ds;
}

Datasource *datasource_new_silent(Dataformat *format, off_t samples)
{
     Datasource *ds;
     ds = gtk_type_new(datasource_get_type());
     memcpy(&(ds->format), format, sizeof(Dataformat));
     ds->type = DATASOURCE_SILENCE;
     ds->length = samples;
     ds->size = samples * ds->format.samplebytes;
     return ds;
}

Datasource *datasource_new_from_data(void *data, Dataformat *format, 
				     guint32 size)
{
     Datasource *ds;
     ds = (Datasource *)gtk_type_new(datasource_get_type());
     ds->type = DATASOURCE_REAL;
     memcpy(&(ds->format),format,sizeof(Dataformat));
     ds->length = size / format->samplebytes;
     ds->size = size;
     ds->data.real = data;
     return ds;
}

#define DUMP_BUFSIZE 65536

gboolean datasource_dump(Datasource *ds, off_t position, 
			 off_t length, EFILE *file, gboolean bigendian,
			 StatusBar *bar)
{
     gchar *buf;
     off_t i;
     guint u;
     if (ds->type == DATASOURCE_BYTESWAP)
	  return datasource_dump(ds->data.clone,position,length,file,
				 !bigendian,bar);
     if (datasource_open(ds)) return TRUE;
     buf = g_malloc(DUMP_BUFSIZE);
     while (length > 0) {
	  i = MIN(length*ds->format.samplebytes,DUMP_BUFSIZE);
	  u = datasource_read_array(ds,position,i,buf);
	  if (XOR(bigendian,IS_BIGENDIAN)) 
	       byteswap(buf,ds->format.samplesize,u);
	  if (u==0 || e_fwrite(buf,u,file) || status_bar_progress(bar,u)) { 
	       datasource_close(ds); 
	       g_free(buf);
	       return TRUE; 
	  }
	  length -= u/ds->format.samplebytes;
	  position += u/ds->format.samplebytes;
     }
     datasource_close(ds);
     g_free(buf);
     return FALSE;
}

gboolean datasource_realize(Datasource *ds)
{
     gchar *c;
     guint32 sz=ds->size;

     if (datasource_open(ds)) return TRUE;

     c = g_malloc(sz);
     if (datasource_read_array(ds,0,sz,c)) {
	  g_free(c);
	  datasource_close(ds);
	  return TRUE;
     }
     datasource_close(ds);
     
     datasource_clear(ds);
     ds->type = DATASOURCE_REAL;
     ds->data.real = c;     

     return FALSE;
}

static gboolean datasource_open_sndfile(Datasource *ds)
{
#if defined(HAVE_LIBSNDFILE)
     SF_INFO i;
     SNDFILE *s;
     char *d;
     int fd;
     fd = xopen(ds->data.sndfile.filename,O_RDONLY,0);
     if (fd == -1) return TRUE;
     s = sf_open_fd(fd,SFM_READ,&i,TRUE);
     if (s == NULL) {	  
	  d = g_strdup_printf("Couldn't open %s",
			      ds->data.sndfile.filename);
	  user_error(d);
	  g_free(d);
	  close(fd);
	  return TRUE;
     }

     ds->data.sndfile.handle = s;
     ds->data.sndfile.pos = 0;

     sf_command ( s, SFC_SET_NORM_FLOAT, NULL, SF_TRUE );
     sf_command ( s, SFC_SET_NORM_DOUBLE, NULL, SF_TRUE );
     return FALSE;
#else
     g_assert_not_reached();
     return TRUE;
#endif     
}

static void datasource_close_sndfile(Datasource *ds)
{
#if defined(HAVE_LIBSNDFILE)
     sf_close(ds->data.sndfile.handle);
#else
     g_assert_not_reached();
#endif
}

gboolean datasource_open(Datasource *ds)
{
     if (ds->opencount == 0) 
	  switch (ds->type) {
	  case DATASOURCE_VIRTUAL:
	  case DATASOURCE_TEMPFILE:
	  case DATASOURCE_TEMPFILE_FLOAT:
	       ds->data.virtual.handle = e_fopen(ds->data.virtual.filename,
						 EFILE_READ);
	       if (ds->data.virtual.handle == NULL) return TRUE;
	       ds->data.virtual.pos = 0;
	       break;
	  case DATASOURCE_SNDFILE:
	  case DATASOURCE_SNDFILE_TEMPORARY:
	       if (datasource_open_sndfile(ds)) return TRUE;
	       break;
	  case DATASOURCE_REF:
	  case DATASOURCE_CLONE:
	  case DATASOURCE_BYTESWAP:	       
	       if (datasource_open(ds->data.clone)) return TRUE;
	  }     
     ds->opencount ++;
     return FALSE;
}

void datasource_close(Datasource *ds)
{
     g_assert(ds->opencount != 0);
     ds->opencount --;
     if (ds->opencount > 0) return;
     switch (ds->type) {
     case DATASOURCE_VIRTUAL:
     case DATASOURCE_TEMPFILE:
     case DATASOURCE_TEMPFILE_FLOAT:
	  e_fclose(ds->data.virtual.handle);
	  break;
     case DATASOURCE_SNDFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
	  datasource_close_sndfile(ds);
	  break;
     case DATASOURCE_REF:
     case DATASOURCE_CLONE:
     case DATASOURCE_BYTESWAP:
	  datasource_close(ds->data.clone);
	  break;
     }     
}

static guint datasource_clone_read_array(Datasource *source, off_t sampleno,
					 guint size, gpointer buffer)
{
     /* This is not optimized but it's only used in rare cases so the important
      * thing is that it works.*/
     off_t orig_offset;
     off_t orig_sampleno;
     off_t orig_size;
     off_t orig_adjust;
     guint x;
     gchar *p;
     /* Offset to be sent to the original chunk */
     orig_offset = sampleno * source->format.samplebytes;
     /* orig_offset converted to sample number */
     orig_sampleno = orig_offset / source->data.clone->format.samplebytes;
     /* How bany bytes too early will we read? */
     orig_adjust = orig_offset % source->data.clone->format.samplebytes;
     /* How much data should we read (should be able to fill buffer) */
     orig_size = size + orig_adjust + source->data.clone->format.samplebytes - 1;
     p = g_malloc(orig_size);
     x = datasource_read_array(source->data.clone, orig_sampleno, orig_size, p);
     if (x != 0) {
	  g_assert(x-orig_adjust >= size);
	  memcpy(buffer, p+orig_adjust, size);
	  x = size;
     }
     g_free(p);
     return x;			       
}

#if defined(HAVE_LIBSNDFILE)
static void sndfile_read_error(sf_count_t x, Datasource *source)
{
     gchar c[256],*d;
     if (x<0) sf_error_str(source->data.sndfile.handle,c,sizeof(c));
     else strcpy(c,"Unexpected end of file");
     d = g_strdup_printf("Error reading %s: %s",
			 source->data.sndfile.filename, c);
     user_error(d);
     g_free(d);
}
#endif

static guint datasource_sndfile_read_array(Datasource *source, 
					   off_t sampleno, guint size,
					   gpointer buffer)
{
#if defined(HAVE_LIBSNDFILE)
     gchar c[256],*d;
     sf_count_t x;
     g_assert(!source->fake_pcm);
     if (source->data.sndfile.pos != sampleno) {
	  if (sf_seek(source->data.sndfile.handle,sampleno,SEEK_SET) == -1) {
	       sf_error_str(source->data.sndfile.handle,c,sizeof(c));
	       d = g_strdup_printf("Error seeking in %s: %s",
				   source->data.sndfile.filename, c);
	       user_error(d);
	       g_free(d);
	       return 0;
	  }
	  source->data.sndfile.pos = sampleno;
     }
     if (source->data.sndfile.raw_readable) {
	  /* Read raw */
	  x = sf_read_raw(source->data.sndfile.handle,buffer,size);	  
	  if (x>0) source->data.sndfile.pos += x/source->format.samplebytes;
	  if (x<size) {
	       sndfile_read_error(x,source);
	       return 0;
	  }
	  return x;
     } else if (source->format.samplesize==2 && source->format.sign) {
	  /* Read as short */
	  x = sf_readf_short(source->data.sndfile.handle,buffer,
			     size / source->format.samplebytes);
	  if (x>0) source->data.sndfile.pos += x;
	  if (x < size / source->format.samplebytes) {
	       sndfile_read_error(x,source);
	       return 0;
	  }
	  return x * source->format.samplebytes;
     } else {	  
	  g_assert_not_reached();
	  return 0;
     } 
#else
     g_assert_not_reached();
     return 0;
#endif
}

static guint datasource_read_fake_pcm(Datasource *source, off_t sampleno,
				      guint size, gpointer buffer)
{
     guint u,x;
     gpointer b;
     u = size / source->format.samplebytes;
     b = g_malloc(u * source->format.channels * sizeof(sample_t));
     x = datasource_read_array_fp(source,sampleno,u,b);
     if (x > 0)
	  convert_float_pcm_array((sample_t *)b,buffer,&(source->format),
				  x*source->format.channels);
     g_free(b);
     return x*source->format.samplebytes;
     
}

static guint datasource_read_array_main(Datasource *source, off_t sampleno, 
					guint size, gpointer buffer)
{
     off_t x;
     guint u;
     if (source->fake_pcm) 
	  return datasource_read_fake_pcm(source,sampleno,size,buffer);

     switch (source->type) {
     case DATASOURCE_REAL:
	  memcpy(buffer,source->data.real+sampleno*source->format.samplebytes,
		 size);
	  return size;
     case DATASOURCE_VIRTUAL: 
     case DATASOURCE_TEMPFILE:
	  /* Calculate offset in file */
	  x = source->data.virtual.offset + 
	       sampleno*source->format.samplebytes;
	  if (x != source->data.virtual.pos && 
	      e_fseek(source->data.virtual.handle, x, SEEK_SET))
	       return 0;
	  source->data.virtual.pos = x;
	  if (e_fread(buffer,size,source->data.virtual.handle)) return 0;
	  source->data.virtual.pos += size;
	  return size;
     case DATASOURCE_SILENCE:
	  memset(buffer,0,size);
	  return size;
     case DATASOURCE_SNDFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
	  return datasource_sndfile_read_array(source,sampleno,size,buffer);
     case DATASOURCE_REF:
	  return datasource_read_array_main(source->data.clone,sampleno,size,
					    buffer);
     case DATASOURCE_CLONE:
	  return datasource_clone_read_array(source,sampleno,size,buffer);
     case DATASOURCE_BYTESWAP:
	  u = datasource_read_array_main(source->data.clone,sampleno,size,
					 buffer);
	  if (u>0) byteswap(buffer,source->format.samplesize,u);
	  return u;
     default:
	  g_assert_not_reached();
	  return 0;
     }
}

guint datasource_read_array(Datasource *source, off_t sampleno, guint size,
			    gpointer buffer)
{
     off_t o;
     g_assert(source->opencount > 0);
     /* Check sampleno */
     g_assert(sampleno <= source->length);     
     /* Check size */
     /* Round down to even samples */     
     size = size - size % source->format.samplebytes; 
     o = (source->length - sampleno) * (off_t)(source->format.samplebytes);
     if (size > o) size = (guint)o; /* Round down to available data */
     if (size == 0) return 0;
     /* Do it */
     return datasource_read_array_main(source,sampleno,size,buffer);
}

gboolean datasource_read(Datasource *source, off_t sampleno, gpointer buffer)
{
     return (datasource_read_array(source,sampleno,source->format.samplebytes,
				   buffer) != source->format.samplebytes);
}

static guint datasource_sndfile_read_array_fp(Datasource *source, 
					      off_t sampleno, guint samples,
					      sample_t *buffer)
{
#if defined(HAVE_LIBSNDFILE)
     gchar c[256],*d;
     sf_count_t x;

     if (source->data.sndfile.pos != sampleno) {
	  if (sf_seek(source->data.sndfile.handle,sampleno,SEEK_SET) == -1) {
	       sf_error_str(source->data.sndfile.handle,c,sizeof(c));
	       d = g_strdup_printf("Error seeking in %s: %s",
				   source->data.sndfile.filename, c);
	       user_error(d);
	       g_free(d);
	       return 0;
	  }
	  source->data.sndfile.pos = sampleno;
     }

     x = sf_readf_sample_t(source->data.sndfile.handle,buffer,samples);
     if (x>0) source->data.sndfile.pos += x;
     if (x<samples) {
	  sndfile_read_error(x,source);
	  return 0;		    
     }
     return x;

#else
     g_assert_not_reached();
     return 0;
#endif
}

guint datasource_read_array_fp(Datasource *source, off_t sampleno,
			       guint samples, sample_t *buffer)
{
     gchar *p;
     guint x,s;
     off_t u;
     sample_t *sp;
     
     g_assert(sampleno <= source->length);
     if (samples > source->length-sampleno) 
	  samples = (guint)(source->length-sampleno);
     if (samples == 0) return 0;

     switch (source->type) {
     case DATASOURCE_REAL_FLOAT:
	  sp = (sample_t *)source->data.real;
	  memcpy(buffer,sp+(sampleno*source->format.channels),
		 samples*source->format.channels*sizeof(sample_t));
	  return samples;
     case DATASOURCE_TEMPFILE_FLOAT:
	  u = source->data.virtual.offset +
	       sampleno*source->format.channels*sizeof(sample_t);
	  if (u != source->data.virtual.pos &&
	      e_fseek(source->data.virtual.handle, u, SEEK_SET))
	       return 0;
	  source->data.virtual.pos = u;
	  if (e_fread(buffer,samples*source->format.channels*sizeof(sample_t),
		      source->data.virtual.handle)) return 0;
	  source->data.virtual.pos += 
	       samples * source->format.channels*sizeof(sample_t);
	  return samples;
     case DATASOURCE_SILENCE:
	  memset(buffer,0,samples*source->format.channels*sizeof(sample_t));
	  return samples;
     case DATASOURCE_SNDFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
       return datasource_sndfile_read_array_fp(source,sampleno,samples,
					       buffer);
     case DATASOURCE_REF:
	  return datasource_read_array_fp(source->data.clone,sampleno,
					  samples,buffer);
     default:
	  g_assert(!source->fake_pcm);
	  s = samples * source->format.samplebytes;
	  p = g_malloc(s);
	  x = datasource_read_array(source,sampleno,s,p);
	  g_assert(x==s || x==0);
	  if (x==s) {
	       convert_pcm_float_array(p,&(source->format),buffer,
				       samples*source->format.channels);
	       g_free(p);
	       return samples;
	  }
	  g_free(p);
	  return 0;
     }
}

gboolean datasource_read_fp(Datasource *ds, off_t sampleno, sample_t *buffer)
{
     return (datasource_read_array_fp(ds,sampleno,1,buffer)!=1);
}

static gboolean datasource_uses_file(Datasource *ds, gchar *filename)
{
     return ((ds->type==DATASOURCE_VIRTUAL && 
	      is_same_file(ds->data.virtual.filename,filename)) ||
	     ((ds->type==DATASOURCE_SNDFILE &&
	       is_same_file(ds->data.sndfile.filename,filename))));
}

gboolean datasource_backup_unlink(gchar *filename)
{
     GList *l,*q=NULL;
     Datasource *ds;
     Datasource *backup=NULL;
     gchar *lastname=filename,*t;
     guint dirnum=0;
     gint i;

     /* Find out which Datasources actually use the file. */
     for (l=datasource_list; l!=NULL; l=l->next) {
	  ds = (Datasource *)l->data;
	  if (datasource_uses_file(ds,filename))
	       q = g_list_append(q,ds);
     }
     
     /* Now iterate through those files. */
     for (l=q; l!=NULL; l=l->next) {
	  ds = (Datasource *)l->data;
	  if (!datasource_uses_file(ds,filename)) continue;

	  /* FIXME: When multiple tempdir support has been added, try moving to
	   * all different partitions before copying */
	  if (backup == NULL) {
	       dirnum = 0;
	       while (1) {
		    t = get_temp_filename(dirnum);
		    if (t == NULL) {
			 dirnum = 0;
			 t = get_temp_filename(0);
			 i = xrename(lastname,t,TRUE);
			 break;
		    }
		    i = xrename(lastname,t,FALSE);
		    if (i!=2) break;
		    dirnum ++;
		    g_free(t);
	       } 
	       if (i) {
		    g_free(t);
		    return TRUE;
	       }		     
	  } else {
	       /* Special case: the same file has been opened many times. */
	       t = get_temp_filename(0);
	       if (errdlg_copyfile(lastname,t)) {
		    g_free(t);
		    return TRUE;
	       }
	  }
		    
	  switch (ds->type) {
	  case DATASOURCE_VIRTUAL:
	       g_free(ds->data.virtual.filename);
	       ds->data.virtual.filename = t;
	       ds->type = DATASOURCE_TEMPFILE;
	       break;
	  case DATASOURCE_SNDFILE:
	       g_free(ds->data.sndfile.filename);
	       ds->data.sndfile.filename = t;
	       ds->type = DATASOURCE_SNDFILE_TEMPORARY;
	       break;
	  default: 
	       g_assert_not_reached(); 
	       break;
	  }	  
	  backup = ds;
	  lastname = t;
     }

     g_list_free(q);
     return xunlink(filename);
}



/* PCM<->FLOAT SAMPLE CONVERSION ROUTINES */
/* These routines could really use some optimizing if anyone feels like it. */

#define STYPE gfloat
#define CPF convert_pcm_float_f
#define CPFA convert_pcm_float_array_f
#define CFP convert_float_pcm_f
#define CFPA convert_float_pcm_array_f
#include "convert_inc.c"
#define STYPE gdouble
#define CPF convert_pcm_float_d
#define CPFA convert_pcm_float_array_d
#define CFP convert_float_pcm_d
#define CFPA convert_float_pcm_array_d
#include "convert_inc.c"

/* Returns the highest value that can be returned from the pcm->float
 * conversion functions for a certain format. Values higher than that or
 * lower than -1.0 will be clipped when converted back. */
sample_t maximum_float_value(Dataformat *format)
{
     if (!format->sign) return 1.0;
     switch (format->samplesize) {
     case 1:
	  return ((sample_t)127.0)/((sample_t)128.0);
     case 2:
	  return ((sample_t)32767.0)/((sample_t)32768.0);
     case 3:
	  return ((sample_t)8388607.0)/((sample_t)8388608.0);
     case 4:
	  return ((sample_t)2147483647.0)/((sample_t)2147483648.0);
     }
     g_assert_not_reached();
     return ((sample_t)1.0);
}

void datasource_conversion_selftest(void)
{
     guint samplesizes[] = { 1, 1, 2, 2, 3, 3, 4, 4 };
     gboolean signs[] = { FALSE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE };
     guchar pcm_buf[256],pcm_buf2[256],pcm_buf3[256];
     sample_t sbuf[32],sbuf2[32],s;
     guint i,j;
     Dataformat fmt;

	  
     fmt.samplesize = 1;
     fmt.sign = TRUE;

     for (i=0; i<256; i++) {
	  pcm_buf[0] = (gchar)i;
	  s = convert_pcm_float(pcm_buf,&fmt);
	  convert_float_pcm(s,pcm_buf+1,&fmt);
	  printf("%d -> %f -> %d\n",pcm_buf[0],s,pcm_buf[1]);
	  
     }

     /* Generate random data in pcm buffer */
     srand(time(0));     
     for (j=0; j<sizeof(pcm_buf); j++) {
	  pcm_buf[j] = rand() % 256;
	  /* printf("%d,",pcm_buf[j]); */
     }
     /* puts(""); */
     
     /* Perform tests */
     for (i=0; i<8; i++) {
	  fmt.samplesize = samplesizes[i];
	  fmt.sign = signs[i];
	  /* Test array funcs. */
	  convert_pcm_float_array(pcm_buf,&fmt,sbuf,32);
	  convert_float_pcm_array(sbuf,pcm_buf2,&fmt,32);
	  if (memcmp(pcm_buf,pcm_buf2,32*fmt.samplesize)) {
	       printf("Self-test failed for samplesize=%d, %s <array>\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	  }
	  /* Test individual funcs. */
	  memset(pcm_buf3,0,sizeof(pcm_buf3));
	  for (j=0; j<32; j++) {
	       s = convert_pcm_float(pcm_buf+j*fmt.samplesize,&fmt);
	       convert_float_pcm(s,pcm_buf3+j*fmt.samplesize,&fmt);
	  }
	  if (memcmp(pcm_buf,pcm_buf3,32*fmt.samplesize)) {
	       printf("Self-test failed for samplesize=%d, %s\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	  }

	  /* Test that the two functions gave the same result */
	  if (memcmp(pcm_buf2,pcm_buf3,32*fmt.samplesize))
	       printf("Self-test mismatch for samplesize=%d, %s\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");	 

	  /* Test extreme values */
	  sbuf[0] = -1.0;
	  sbuf[1] = maximum_float_value(&fmt);
	  convert_float_pcm_array(sbuf,pcm_buf,&fmt,2);
	  convert_pcm_float_array(pcm_buf,&fmt,sbuf2,2);
	  if (memcmp(sbuf,sbuf2,2*sizeof(sample_t))) {
	       printf("Self-test failed for samplesize=%d, %s <array extremes>\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	       printf("    (%f,%f) => (%f,%f)\n",sbuf[0],sbuf[1],sbuf2[0],sbuf2[1]);
	  }
	  for (j=0; j<2; j++) {
	       convert_float_pcm(sbuf[j],pcm_buf2,&fmt);
	       sbuf2[j] = convert_pcm_float(pcm_buf2,&fmt);
	  }
	  if (memcmp(sbuf,sbuf2,2*sizeof(sample_t))) {
	       printf("Self-test failed for samplesize=%d, %s <extremes>\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	       printf("    (%f,%f) => (%f,%f)\n",sbuf[0],sbuf[1],sbuf2[0],sbuf2[1]);
	  }

     }
}

gboolean dataformat_equal(Dataformat *f1, Dataformat *f2)
{
	return (f1->samplerate==f2->samplerate && f1->samplesize==f2->samplesize && f1->channels==f2->channels && f1->sign==f2->sign);
}

Datasource *datasource_try_convert_samplesize(Datasource *ds, guint ssize, 
					      gboolean sign)
{
     Dataformat tfmt;
     Datasource *r;
     if (!ds->fake_pcm) return NULL;
     memcpy(&tfmt,&(ds->format),sizeof(Dataformat));
     tfmt.samplesize = ssize;
     tfmt.sign = sign;
     tfmt.samplebytes = tfmt.samplesize * tfmt.channels;
     r = datasource_copy(ds);
     g_assert(r->fake_pcm);
     memcpy(&(r->format),&tfmt,sizeof(Dataformat));
     r->size = r->length * r->format.samplebytes;
     return r;

     
}
